@@ -129,6 +129,17 @@ static drmServerInfo drm_server_info = {
 	dri_drm_get_perms,
 };
 
+static void
+free_msm(MSMPtr pMsm)
+{
+	if (pMsm->drmFD)
+#ifdef XF86_PDEV_SERVER_FD
+		if (!(pMsm->pEnt->location.type == BUS_PLATFORM &&
+			(pMsm->pEnt->location.id.plat->flags & XF86_PDEV_SERVER_FD)))
+#endif
+		drmClose(pMsm->drmFD);
+	free(pMsm);
+}
 
 static Bool
 MSMInitDRM(ScrnInfoPtr pScrn)
@@ -212,7 +223,7 @@ MSMPreInit(ScrnInfoPtr pScrn, int flags)
 	pScrn->progClock = TRUE;
 	pScrn->chipset = MSM_DRIVER_NAME;
 
-	INFO_MSG("MSM/Qualcomm processor (video memory: %dkB)", pScrn->videoRam / 1024);
+	INFO_MSG("MSM/Qualcomm processor");
 
 	if (!MSMInitDRM(pScrn)) {
 		ERROR_MSG("Unable to open DRM");
@@ -236,7 +247,7 @@ MSMPreInit(ScrnInfoPtr pScrn, int flags)
 	pMsm->options = malloc(sizeof(MSMOptions));
 
 	if (pMsm->options == NULL) {
-		free(pMsm);
+		free_msm(pMsm);
 		return FALSE;
 	}
 
@@ -279,18 +290,18 @@ MSMPreInit(ScrnInfoPtr pScrn, int flags)
 	xf86SetDpi(pScrn, 0, 0);
 
 	if (!xf86SetWeight(pScrn, defaultWeight, defaultWeight)) {
-		free(pMsm);
+		free_msm(pMsm);
 		return FALSE;
 	}
 
 	/* Initialize default visual */
 	if (!xf86SetDefaultVisual(pScrn, -1)) {
-		free(pMsm);
+		free_msm(pMsm);
 		return FALSE;
 	}
 
 	if (!xf86SetGamma(pScrn, zeros)) {
-		free(pMsm);
+		free_msm(pMsm);
 		return FALSE;
 	}
 
@@ -545,6 +556,14 @@ MSMLeaveVT(VT_FUNC_ARGS_DECL)
 	}
 }
 
+static void
+MSMFreeScreen(FREE_SCREEN_ARGS_DECL)
+{
+	SCRN_INFO_PTR(arg);
+	MSMPtr pMsm = MSMPTR(pScrn);
+	free_msm(pMsm);
+}
+
 /* ------------------------------------------------------------ */
 /* Following is the standard driver setup that probes for the   */
 /* hardware and sets up the structures.                         */
@@ -657,12 +676,113 @@ MSMProbe(DriverPtr drv, int flags)
 		pScrn->SwitchMode = MSMSwitchMode;
 		pScrn->EnterVT = MSMEnterVT;
 		pScrn->LeaveVT = MSMLeaveVT;
+		pScrn->FreeScreen = MSMFreeScreen;
 	}
 
 	free(sections);
 	return foundScreen;
 }
 
+static Bool
+MSMDriverFunc(ScrnInfoPtr scrn, xorgDriverFuncOp op, void *data)
+{
+	xorgHWFlags *flag;
+
+	switch (op) {
+	case GET_REQUIRED_HW_INTERFACES:
+		flag = (CARD32 *)data;
+		(*flag) = 0;
+		return TRUE;
+#ifdef XF86_PDEV_SERVER_FD
+	case SUPPORTS_SERVER_FDS:
+		return TRUE;
+#endif
+	default:
+		return FALSE;
+	}
+}
+
+#ifdef XSERVER_PLATFORM_BUS
+static Bool probe_hw(struct xf86_platform_device *dev)
+{
+	int fd;
+
+	/* NOTE: for kgsl we still need config file to find fbdev device,
+	 * so it will always be probed through the old MSMProbe path.  So
+	 * only look for drm/msm here:
+	 */
+
+#if XF86_PDEV_SERVER_FD
+	if (dev && (dev->flags & XF86_PDEV_SERVER_FD)) {
+		drmVersionPtr version;
+
+		fd = xf86_get_platform_device_int_attrib(dev, ODEV_ATTRIB_FD, -1);
+		if (fd == -1)
+			return FALSE;
+
+		version = drmGetVersion(fd);
+		/* make sure we have the right device: */
+		if (version && (strcmp(version->name, "msm") == 0)) {
+			drmFreeVersion(version);
+			return TRUE;
+		}
+
+		drmFreeVersion(version);
+		return FALSE;
+	}
+#endif
+
+	fd = drmOpen("msm", NULL);
+	if (fd != -1) {
+		close(fd);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static Bool
+MSMPlatformProbe(DriverPtr driver,
+		int entity_num, int flags, struct xf86_platform_device *dev,
+		intptr_t match_data)
+{
+	ScrnInfoPtr pScrn = NULL;
+	int scr_flags = 0;
+
+	/* Note: at least for now there is no point in gpu screens.. and
+	 * allowing them exposes a bug in older xservers that would result
+	 * in the device probed as a gpu screen rather than regular screen
+	 * resulting in "No screens found".
+	 *
+	 * If later there is actually reason to support GPU screens, track
+	 * down the first xorg ABI # that contains the fix, and make this
+	 * conditional on that or later ABI versions.  Otherwise you will
+	 * break things for people with older xservers.
+	 *
+	if (flags & PLATFORM_PROBE_GPU_SCREEN)
+		scr_flags = XF86_ALLOCATE_GPU_SCREEN;
+	 */
+
+	if (probe_hw(dev)) {
+		pScrn = xf86AllocateScreen(driver, scr_flags);
+		xf86AddEntityToScreen(pScrn, entity_num);
+
+		pScrn->driverVersion = MSM_VERSION_CURRENT;
+		pScrn->driverName = MSM_NAME;
+		pScrn->name = MSM_NAME;
+		pScrn->Probe = MSMProbe;
+		pScrn->PreInit = MSMPreInit;
+		pScrn->ScreenInit = MSMScreenInit;
+		pScrn->SwitchMode = MSMSwitchMode;
+		pScrn->EnterVT = MSMEnterVT;
+		pScrn->LeaveVT = MSMLeaveVT;
+		pScrn->FreeScreen = MSMFreeScreen;
+	}
+
+	return pScrn != NULL;
+}
+#endif
+
 _X_EXPORT DriverRec freedrenoDriver = {
 		MSM_VERSION_CURRENT,
 		MSM_DRIVER_NAME,
@@ -671,7 +791,12 @@ _X_EXPORT DriverRec freedrenoDriver = {
 		MSMAvailableOptions,
 		NULL,
 		0,
-		NULL
+		MSMDriverFunc,
+		NULL,
+		NULL,  /* pci_probe */
+#ifdef XSERVER_PLATFORM_BUS
+		MSMPlatformProbe,
+#endif
 };
 
 MODULESETUPPROTO(freedrenoSetup);
